#pragma once

#include "proto/ProtoDisableWarning.h"

#include <string>

#include "InworldState.h"

#include "HAL/RunnableThread.h"
#include "HAL/Runnable.h"
#include "Containers/Queue.h"

#if PLATFORM_WINDOWS
#include "Windows/PreWindowsApi.h"
#endif
#include "proto/world-engine.grpc.pb.h"
#if PLATFORM_WINDOWS
#include "Windows/PostWindowsApi.h"
#endif

#include "proto/ai/inworld/studio/v1alpha/tokens.grpc.pb.h"
#include "proto/ai/inworld/studio/v1alpha/tokens.pb.h"
#include "proto/packets.pb.h"

namespace InworldEngine = ai::inworld::engine;
namespace InworldPackets = ai::inworld::packets;
namespace InworldV1alpha = ai::inworld::studio::v1alpha;

using grpc::Channel;
using grpc::ClientContext;
using grpc::Status;

using FReaderWriter = ::grpc::ClientReaderWriter< InworldPackets::InworldPacket, InworldPackets::InworldPacket>;

namespace Inworld
{
	class FInworldPacket;
	class FTextEvent;
	class FDataEvent;
	class FCustomEvent;

	class INWORLDAICLIENT_API FInworldRunnable : public FRunnable
	{
	public:
		bool IsDone() const { return bDone; };
		virtual void Stop() override { bDone = true; }

	protected:
		TAtomic<bool> bDone = false;
	};

    class INWORLDAICLIENT_API FAsyncRoutine
    {
    public:
		~FAsyncRoutine() { Stop(); }

		void Start(FInworldRunnable* InRunnable, const TCHAR* ThreadName);
		void Stop();

		bool IsValid() { return Runnable && Thread; }
        bool IsDone() { return Runnable && Runnable->IsDone(); }

        template<class T>
        T* GetTask() { return static_cast<T*>(Runnable); }

    private:
        FRunnableThread* Thread = nullptr;
		FInworldRunnable* Runnable = nullptr;
    };

    class INWORLDAICLIENT_API FRunnableMessaging : public FInworldRunnable
    {
	public:
        FRunnableMessaging(FReaderWriter& InReaderWriter, TQueue<TSharedPtr<Inworld::FInworldPacket>>& InPackets, TAtomic<bool>& InIsError)
			: ReaderWriter(InReaderWriter)
			, Packets(InPackets)
			, bIsError(InIsError)
		{}

	protected:
		FReaderWriter& ReaderWriter;
		TQueue<TSharedPtr<Inworld::FInworldPacket>>& Packets;
		TAtomic<bool>& bIsError;
    };

    class INWORLDAICLIENT_API FRunnableRead : public FRunnableMessaging
    {
    public:
        FRunnableRead(FReaderWriter& InReaderWriter, TQueue<TSharedPtr<Inworld::FInworldPacket>>& InPackets, TAtomic<bool>& InIsError)
			: FRunnableMessaging(InReaderWriter, InPackets, InIsError)
		{}

        virtual uint32 Run() override;
	};

	class FRunnableWrite : public FRunnableMessaging
	{
	public:
        FRunnableWrite(FReaderWriter& InReaderWriter, TQueue<TSharedPtr<Inworld::FInworldPacket>>& InPackets, TAtomic<bool>& InIsError)
			: FRunnableMessaging(InReaderWriter, InPackets, InIsError)
		{}

		virtual uint32 Run() override;
	};

	class INWORLDAICLIENT_API FRunnableLoadScene : public FInworldRunnable
	{
	public:
        FRunnableLoadScene(const FString& InAuthUrl, const FString& InLoadSceneUrl, const FString& InSceneName, const FString& InApiKey, const FString& InApiSecret, const FString& InPlayerName)
			: AuthUrl(InAuthUrl)
			, LoadSceneUrl(InLoadSceneUrl)
			, SceneName(InSceneName)
			, ApiKey(InApiKey)
			, ApiSecret(InApiSecret)
			, PlayerName(InPlayerName)
		{}

		virtual uint32 Run() override;
		virtual void Stop() override;
		
        std::unique_ptr<FReaderWriter> Session();

		grpc::Status& GetStatus() { return Status; }
		InworldEngine::LoadSceneResponse& GetResponse() { return Response; }

		std::unique_ptr<ClientContext>& GetContext() { return Context; }

	private:
		std::string GenerateHeader() const;
		std::string GenerateUserId() const;

		std::unique_ptr<InworldEngine::WorldEngine::Stub> Stub;

		FString AuthUrl;
		FString LoadSceneUrl;
		FString SceneName;
		FString ApiKey;
		FString ApiSecret;
		FString PlayerName;

		std::unique_ptr<ClientContext> Context;
		InworldV1alpha::SessionAccessToken Token;
		grpc::Status Status;
		InworldEngine::LoadSceneResponse Response;
	};

    class INWORLDAICLIENT_API FAsyncRoutineMessaging
    {
    public:
        FAsyncRoutineMessaging(FReaderWriter& InReaderWriter, TQueue<TSharedPtr<Inworld::FInworldPacket>>& InPackets, TAtomic<bool>& InIsError, TAtomic<bool>& InQuitRoutines)
            : ReaderWriter(InReaderWriter)
            , Packets(InPackets)
            , bIsError(InIsError)
        {}

        bool CanAbandon() { return true; }

        void Abandon() { }

        FORCEINLINE TStatId GetStatId() const
        {
            RETURN_QUICK_DECLARE_CYCLE_STAT(FAsyncRoutineMessaging, STATGROUP_ThreadPoolAsyncTasks);
        }

    protected:
        FReaderWriter& ReaderWriter;
        TQueue<TSharedPtr<Inworld::FInworldPacket>>& Packets;
        TAtomic<bool>& bIsError;
	};

	struct FAgentInfo
	{
		FString BrainName;
		FName AgentId;
	};

	class INWORLDAICLIENT_API FClient
	{
	public:
		void SendPacket(TSharedPtr<Inworld::FInworldPacket> Packet);
		TSharedRef<FTextEvent> SendTextMessage(const FName& AgentId, const FString& Text);
		TSharedRef<FDataEvent> SendSoundMessage(const FName& AgentId, const std::string& Data);
		TSharedRef<FCustomEvent> SendCustomEvent(FName AgentId, const FString& Name);

		void CancelResponse(const FName& AgentId, const FName& InteractionId, const TArray<FName>& UtteranceIds);

		void StartAudioSession(const FName& AgentId);
		void StopAudioSession(const FName& AgentId);

		void CreateClient(const FString& AuthUrl, const FString& LoadSceneUrl, const FString& SceneName, const FString& ApiKey, const FString& PlayerName, const FString& ApiSecret, TFunction<void(const TArray<Inworld::FAgentInfo>&)> LoadSceneCallback, TFunction<void(EInworldConnectionState ConnectionState)> ConnectionStateCallback);
		void DestroyClient();

		TQueue<TSharedPtr<Inworld::FInworldPacket>>& GetIncomingPackets() { return IncomingPackets; }
		EInworldConnectionState GetConnectionState() const { return ConnectionState; }

		void Tick(float DeltaTime);

	private:
		bool ProcessLoadScene();

		void ConnectToSession();

		void SetConnectionState(EInworldConnectionState State);

		void CancelTasks();

		TFunction<void(const TArray<Inworld::FAgentInfo>&)> OnLoadSceneCallback;
		TFunction<void(EInworldConnectionState ConnectionState)> OnConnectionStateChangedCallback;

		std::unique_ptr<FReaderWriter> ReaderWriter;

		FAsyncRoutine AsyncReadTask;
        FAsyncRoutine AsyncWriteTask;
        FAsyncRoutine AsyncLoadSceneTask;

		TQueue<TSharedPtr<FInworldPacket>> IncomingPackets;
		TQueue<TSharedPtr<FInworldPacket>> OutgoingPackets;

		TAtomic<bool> bIsError = false;

		EInworldConnectionState ConnectionState = EInworldConnectionState::Idle;
	};
}