#pragma once

#include "InworldApi.h"
#include "proto/ProtoDisableWarning.h"
#include "Packets.h"
#include "InworldComponentInterface.h"
#include "InworldUtils.h"

void UInworldApiSubsystem::StartSession(const FString& SceneName, const FString& PlayerName)
{
	if (AuthUrl.IsEmpty())
	{
		Inworld::Utils::ErrorLog("AuthUrl is empty. Make sure you've set up [/Script/InworldAIIntegration.InworldApiSubsystem] AuthUrl in Engine.ini.", true);
		return;
	}
	if (TargetUrl.IsEmpty())
	{
		Inworld::Utils::ErrorLog("TargetUrl is empty. Make sure you've set up [/Script/InworldAIIntegration.InworldApiSubsystem] TargetUrl in Engine.ini.", true);
		return;
	}
	if (ApiKey.IsEmpty())
	{
		Inworld::Utils::ErrorLog("ApiKey is empty. Make sure you've set up [/Script/InworldAIIntegration.InworldApiSubsystem] ApiKey in Engine.ini.", true);
		return;
	}
    if (ApiSecret.IsEmpty())
    {
        Inworld::Utils::ErrorLog("ApiSecret is empty. Make sure you've set up [/Script/InworldAIIntegration.InworldApiSubsystem] ApiSecret in Engine.ini.", true);
        return;
    }

    Inworld::Utils::PrepareSslCreds();

    Client.CreateClient(AuthUrl, TargetUrl, SceneName, ApiKey, ApiSecret, PlayerName,
        [this](const auto& AgentInfos)
        {
            for (const auto& Info : AgentInfos)
            {
                auto* ComponentPtr = CharacterComponentRegistry.FindByPredicate([&Info](const auto C) { return C->GetBrainName() == Info.BrainName; });
                if (ComponentPtr)
                {
					auto& Component = *ComponentPtr;
					CharacterComponentByAgentId.Add(Info.AgentId, Component);
					Component->SetAgentId(Info.AgentId);
                }
            }
        },
        [this](EInworldConnectionState ConnectionState)
        {
            if (ensure(PlayerComponent))
            {
                PlayerComponent->HandleConnectionStateChanged(ConnectionState);
            }

            OnConnectionStateChanged.Broadcast(ConnectionState);
        });
}

void UInworldApiSubsystem::RegisterCharacterComponent(Inworld::ICharacterComponent* Component)
{
    CharacterComponentRegistry.Add(Component);
}

void UInworldApiSubsystem::UnregisterCharacterComponent(Inworld::ICharacterComponent* Component)
{
    CharacterComponentRegistry.Remove(Component);
    CharacterComponentByAgentId.Remove(Component->GetAgentId());
}

void UInworldApiSubsystem::RegisterPlayerComponent(Inworld::IPlayerComponent* Component)
{
    PlayerComponent = Component;
}

void UInworldApiSubsystem::UnregisterPlayerComponent()
{
    PlayerComponent = nullptr;
}

void UInworldApiSubsystem::SendTextMessage(const FName& AgentId, const FString& Text)
{
    auto Packet = Client.SendTextMessage(AgentId, Text);
    if (PlayerComponent)
    {
        PlayerComponent->HandlePacket(Packet);
    }
}

void UInworldApiSubsystem::SendCustomEvent(FName AgentId, const FString& Name)
{
    Client.SendCustomEvent(AgentId, Name);
}

void UInworldApiSubsystem::SendCustomEventForTargetAgent(const FString& Name)
{
    if (ensure(PlayerComponent))
    {
        if (auto* Character = PlayerComponent->GetTargetCharacter())
        {
            SendCustomEvent(Character->GetAgentId(), Name);
        }
    }
}

void UInworldApiSubsystem::SendAudioMessage(const FName& AgentId, USoundWave* SoundWave)
{
    std::string Data;
    if (ensure(Inworld::Utils::SoundWaveToString(SoundWave, Data)))
    {
        Client.SendSoundMessage(AgentId, Data);
    }
}

void UInworldApiSubsystem::StartAudioSession(const FName& AgentId)
{
    Client.StartAudioSession(AgentId);
}

void UInworldApiSubsystem::StopAudioSession(const FName& AgentId)
{
    Client.StopAudioSession(AgentId);
}

void UInworldApiSubsystem::CancelResponse(const FName& AgentId, const FName& InteractionId, const TArray<FName>& UtteranceIds)
{
    Client.CancelResponse(AgentId, InteractionId, UtteranceIds);
}

bool UInworldApiSubsystem::DoesSupportWorldType(EWorldType::Type WorldType) const
{
    return WorldType == EWorldType::Game || WorldType == EWorldType::PIE;
}

void UInworldApiSubsystem::Tick(float DeltaTime)
{
    Super::Tick(DeltaTime);

    auto& Packets = Client.GetIncomingPackets();
    TSharedPtr<Inworld::FInworldPacket> Packet;
    while (Packets.Dequeue(Packet))
    {
        if (!Packet)
        {
            continue;
        }

        auto* ComponentPtr = CharacterComponentByAgentId.Find(Packet->Routing.Source.Name);
        if (ComponentPtr)
        {
            (*ComponentPtr)->HandlePacket(Packet);
        }
        else if (Packet->Routing.Source.Type == ai::inworld::packets::Actor_Type_PLAYER && PlayerComponent)
        {
            PlayerComponent->HandlePacket(Packet);
        }
    }

    Client.Tick(DeltaTime);
}

void UInworldApiSubsystem::Deinitialize()
{
    Client.DestroyClient();
}

