#include "InworldCharacterComponent.h"
#include "proto/ProtoDisableWarning.h"
#include "InworldApi.h"
#include "Packets.h"
#include "Components/AudioComponent.h"
#if INWORLD_OVR_LIPSYNC
#include "OVRLipSyncFrame.h"
#include "OVRLipSyncContextWrapper.h"
#include "OVRLipSync.h"
#include "OVRLipSyncPlaybackActorComponent.h"
#endif
#include "Engine/EngineBaseTypes.h"
#include "Animation/AnimMontage.h"
#include "InworldUtils.h"

static TAutoConsoleVariable<bool> CVarSkipCharacterAudio(
	TEXT("Inworld.SkipCharacterAudio"),
	false,
	TEXT(""));

UInworldCharacterComponent::UInworldCharacterComponent()
{
    PrimaryComponentTick.bCanEverTick = true;
}

void UInworldCharacterComponent::BeginPlay()
{
    Super::BeginPlay();

    AudioComponent = Cast<UAudioComponent>(GetOwner()->GetComponentByClass(UAudioComponent::StaticClass()));

#if INWORLD_OVR_LIPSYNC
    LipSyncComponent = Cast<UOVRLipSyncPlaybackActorComponent>(GetOwner()->GetComponentByClass(UOVRLipSyncPlaybackActorComponent::StaticClass()));
#endif

    if (AudioComponent.IsValid())
    {
        AudioPlaybackPercentHandle = AudioComponent->OnAudioPlaybackPercentNative.AddUObject(this, &UInworldCharacterComponent::OnAudioPlaybackPercent);
    }

    auto* InworldSubsystem = GetWorld()->GetSubsystem<UInworldApiSubsystem>();
    if (ensure(InworldSubsystem))
    { 
        InworldSubsystem->RegisterCharacterComponent(this);
    }
}

void UInworldCharacterComponent::EndPlay(EEndPlayReason::Type Reason)
{
    if (AudioComponent.IsValid())
    {
        AudioComponent->OnAudioPlaybackPercentNative.Remove(AudioPlaybackPercentHandle);
    }

    auto* InworldSubsystem = GetWorld()->GetSubsystem<UInworldApiSubsystem>();
    if (InworldSubsystem)
    {
        InworldSubsystem->UnregisterCharacterComponent(this);
    }

    Super::EndPlay(Reason);
}

void UInworldCharacterComponent::TickComponent(float DeltaTime, enum ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
    Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

    bool bTalking = false;
    if (auto* Audio = AudioComponent.Get())
    {
        if (!Audio->IsPlaying() && PendingUtterances.Num() > 0)
        {
            FInworldCharacterUtterance& Utterance = PendingUtterances[0];
            
            if (Utterance.IsValid())
            {
                float Duration = 0.f;
                if (Utterance.HasTextAndAudio())
                {
                    if (!CVarSkipCharacterAudio.GetValueOnGameThread())
                    {
                        Duration = Utterance.SoundWave->GetDuration();
                        Audio->SetSound(Utterance.SoundWave);
#if INWORLD_OVR_LIPSYNC
                        if (auto* LipSync = LipSyncComponent.Get())
                        {
                            LipSync->Stop();
                            LipSync->Start(Audio, Cast<UOVRLipSyncFrameSequence>(Utterance.LipSyncSequence));
                        }
                        else
#endif
                        {
                            Audio->Play();
                        }
                    }

					PlayerInteractionHistory.Add(Utterance.TextEvent);
					OnInteractionsChanged.Broadcast(PlayerInteractionHistory.GetInteractions());

					CurrentInteractionId = Utterance.TextEvent.PacketId.InteractionId;
					CurrentUtteranceId = Utterance.TextEvent.PacketId.UtteranceId;
				}

				OnUtteranceStarted.Broadcast(Duration, Utterance.CustomGesture);

                PendingUtterances.RemoveAt(0);
            }
        }

        bTalking = Audio->IsPlaying();
        if (bTalking)
        {
            bPendingResponseToPLayer = false;
        }
        else if (CurrentUtteranceId != NAME_None)
		{
			OnUtteranceStopped.Broadcast();
			CurrentUtteranceId = NAME_None;
		}
    }

    EInworldCharacterState State = EInworldCharacterState::Idle;
    if (bTalking)
    {
        State = EInworldCharacterState::Talking;
    }
    else if (bListeningToPlayer)
    {
        State = EInworldCharacterState::Listening;
    }
    else if (bPendingResponseToPLayer)
    {
        State = EInworldCharacterState::Thinking;
    }
    else if (bInteractingWithPlayer)
    {
        State = EInworldCharacterState::Interested;
    }
    SetState(State);
}

void UInworldCharacterComponent::SendTextMessage(const FString& Text)
{
    if (ensure(!AgentId.IsNone()))
    {
        GetWorld()->GetSubsystem<UInworldApiSubsystem>()->SendTextMessage(AgentId, Text);
    }
}

void UInworldCharacterComponent::HandlePacket(TSharedPtr<Inworld::FInworldPacket> Packet)
{
    if (ensure(Packet))
	{
		Packet->Accept(*this);
    }
}

void UInworldCharacterComponent::HandlePlayerInteraction(bool bInteracting)
{
    bInteractingWithPlayer = bInteracting;
    OnPlayerInteractionStateChanged.Broadcast(bInteractingWithPlayer);
}

void UInworldCharacterComponent::HandlePlayerTalking(const Inworld::FTextEvent& Event)
{
    if (CurrentInteractionId != NAME_None && CurrentUtteranceId != NAME_None && CurrentInteractionId != Event.PacketId.InteractionId)
    {
        if (auto* Audio = AudioComponent.Get())
        {
            Audio->Stop();
            OnUtteranceInterrupted.Broadcast();
        }

        PlayerInteractionHistory.CancelUtterance(CurrentInteractionId, CurrentUtteranceId);
        
        TArray<FName> Utterances;
        Utterances.Reserve(PendingUtterances.Num());
        for (auto& Utterance : PendingUtterances)
        {
            if (ensure(CurrentInteractionId == Utterance.InteractionId))
            {
                Utterances.Add(Utterance.UtteranceId);
            }
        }

        if (Utterances.Num() > 0)
        {
            GetWorld()->GetSubsystem<UInworldApiSubsystem>()->CancelResponse(AgentId, CurrentInteractionId, Utterances);
        }

        PendingUtterances.Empty();
    }

    if (!Event.IsFinal())
    {
        bListeningToPlayer = true;
        bPendingResponseToPLayer = true;
    }
    else
    {
        bListeningToPlayer = false;
    }
    Inworld::Utils::Log(FString::Printf(TEXT("%s: %s"), *UiName, *Event.GetText()), true, Inworld::Utils::FTextColors::Player());

    PlayerInteractionHistory.Add(Event);
    OnInteractionsChanged.Broadcast(PlayerInteractionHistory.GetInteractions());
}

template<class TKey, class TDataTable>
UAnimMontage* GetMontageByKey(TKey Key, float UtteranceDuration, const UDataTable* DataTable, TArray<UAnimMontage*>& Montages)
{
	if (!ensure(DataTable))
	{
		Inworld::Utils::ErrorLog("Setup animation data table!");
		return nullptr;
	}

	if (Montages.Num() == 0)
	{
        DataTable->ForeachRow<TDataTable>(TEXT("InworldAnim"),
			[&Montages, Key](const FName& Name, const TDataTable& Row)
			{
				if (Key == Row.Key && ensure(Row.Montages.Num() > 0))
				{
                    Montages = Row.Montages;
				}
			});
	}

	if (Montages.Num() == 0)
	{
		Inworld::Utils::ErrorLog("Setup animation data table!");
		return nullptr;
	}

	int32 Idx = 0;
	float DurationDif = TNumericLimits<float>::Max();
	for (int32 i = 0; i < Montages.Num(); i++)
	{
		auto* Montage = Montages[i];
		const float MontageDuration = Montage->GetPlayLength();
		const float Dif = FMath::Abs(UtteranceDuration - MontageDuration);
		if (Dif < DurationDif)
		{
			Idx = i;
			DurationDif = Dif;
		}
	}

	UAnimMontage* Montage = Montages[Idx];
    Montages.RemoveAt(Idx);

	return Montage;
}

UAnimMontage* UInworldCharacterComponent::GetMontageForEmotion(EInworldCharacterEmotionalBehavior State, float UtteranceDuration) const
{
    auto* Montage = GetMontageByKey<EInworldCharacterEmotionalBehavior, FInworldAnimationTableRow>(State, UtteranceDuration, AnimationDT, CurrentEmotionMontages);
    if (Montage)
    {
        return Montage;
    }

    CurrentEmotionMontages.Empty();
    return GetMontageByKey<EInworldCharacterEmotionalBehavior, FInworldAnimationTableRow>(EInworldCharacterEmotionalBehavior::NEUTRAL, UtteranceDuration, AnimationDT, CurrentEmotionMontages);
}

UAnimMontage* UInworldCharacterComponent::GetMontageForCustomGesture(const FString& Semantic, float UtteranceDuration) const
{
    TArray<UAnimMontage*> Montages;
	return GetMontageByKey<FString, FInworldSemanticGestureTableRow>(Semantic, UtteranceDuration, SemanticGestureDT, Montages);
}

float UInworldCharacterComponent::GetRemainingTimeForCurrentUtterance() const
{
    if (CurrentUtteranceId.IsNone())
    {
        return 0.f;
    }

    auto* Audio = AudioComponent.Get();
    if (!Audio || !Audio->Sound)
    {
        return 0.f;
    }

    return (1.f - CurrentAudioPlaybackPercent) * Audio->Sound->Duration;
}

void UInworldCharacterComponent::SendCustomEvent(const FString& Name) const
{
    GetWorld()->GetSubsystem<UInworldApiSubsystem>()->SendCustomEvent(AgentId, Name);
}

void UInworldCharacterComponent::SetState(EInworldCharacterState State)
{
    if (CurrentState != State)
    {
        CurrentState = State;
        OnStateChanged.Broadcast(CurrentState);
    }
}

FInworldCharacterUtterance* UInworldCharacterComponent::FindOrAddUtterance(const FName& InteractionId, const FName& UtteranceId)
{
    if (PlayerInteractionHistory.IsInteractionCanceled(InteractionId))
    {
        GetWorld()->GetSubsystem<UInworldApiSubsystem>()->CancelResponse(AgentId, InteractionId, { UtteranceId });
        return nullptr;
    }

    if (auto* Utterance = PendingUtterances.FindByPredicate([&InteractionId, &UtteranceId](const auto& U) { return U.InteractionId == InteractionId && U.UtteranceId == UtteranceId; }))
    {
        return Utterance;
    }

    auto& Utterance = PendingUtterances.Emplace_GetRef();
    Utterance.InteractionId = InteractionId;
    Utterance.UtteranceId = UtteranceId;
    return &Utterance;
}

void UInworldCharacterComponent::OnAudioPlaybackPercent(const UAudioComponent* InAudioComponent, const USoundWave* InSoundWave, float Percent)
{
    CurrentAudioPlaybackPercent = Percent;
}

bool UInworldCharacterComponent::IsCustomGesture(const FString& CustomEventName) const
{
    return CustomEventName.Find("gesture") == 0;
}

#if INWORLD_OVR_LIPSYNC
UOVRLipSyncFrameSequence* UInworldCharacterComponent::SoundWaveToLipSyncSequence(USoundWave* SoundWave, uint8* Data, int32 Num)
{
	constexpr float SequenceDuration = 0.01f;

	const int32 NumChannels = SoundWave->NumChannels;
	const int32 SampleRate = SoundWave->GetSampleRateForCurrentPlatform();
	const int32 PCMDataSize = Num / sizeof(int16);
	const int16* PCMData = reinterpret_cast<int16*>(Data);
	const int32 ChunkSizeSamples = static_cast<int32>(SampleRate * SequenceDuration);
	const int32 ChunkSize = NumChannels * ChunkSizeSamples;

	float LaughterScore = 0.f;
	int32 FrameDelayInMs = 0;
	TArray<float> Visemes;

	UOVRLipSyncContextWrapper Context(ovrLipSyncContextProvider_Enhanced, SampleRate, 4096, FString());
	TArray<int16> Samples;
	Samples.SetNumZeroed(ChunkSize);
	Context.ProcessFrame(Samples.GetData(), ChunkSizeSamples, Visemes, LaughterScore, FrameDelayInMs, NumChannels > 1);

	const int32 FrameOffset = static_cast<int32>(FrameDelayInMs * SampleRate / 1000 * NumChannels);

	auto Sequence = NewObject<UOVRLipSyncFrameSequence>(UOVRLipSyncFrameSequence::StaticClass());
	for (int32 Offset = 0; Offset < PCMDataSize + FrameOffset; Offset += ChunkSize)
	{
		const int32 RemainingSamples = PCMDataSize - Offset;
		if (RemainingSamples >= ChunkSize)
		{
			Context.ProcessFrame(PCMData + Offset, ChunkSizeSamples, Visemes, LaughterScore, FrameDelayInMs, NumChannels > 1);
		}
		else
		{
			if (RemainingSamples > 0)
			{
				FMemory::Memcpy(Samples.GetData(), PCMData + Offset, sizeof(int16) * RemainingSamples);
				FMemory::Memset(Samples.GetData() + RemainingSamples, 0, ChunkSize - RemainingSamples);
			}
			else
			{
				FMemory::Memset(Samples.GetData(), 0, ChunkSize);
			}
			Context.ProcessFrame(Samples.GetData(), ChunkSizeSamples, Visemes, LaughterScore, FrameDelayInMs, NumChannels > 1);
		}

		if (Offset >= FrameOffset)
		{
			Sequence->Add(Visemes, LaughterScore);
		}
	}

	return Sequence;
}
#endif

void UInworldCharacterComponent::Visit(const Inworld::FTextEvent& Event)
{
    if (Event.IsFinal())
    {
        Inworld::Utils::Log(FString::Printf(TEXT("%s: %s"), *UiName, *Event.GetText()), true, Inworld::Utils::FTextColors::Character());
    }

    if (auto* Utterance = FindOrAddUtterance(Event.PacketId.InteractionId, Event.PacketId.UtteranceId))
    {
        Utterance->TextEvent = Event;
    }
}

void UInworldCharacterComponent::Visit(const Inworld::FDataEvent& Event)
{
    const auto& Data = Event.GetDataChunk();
    USoundWave* SoundWave = Inworld::Utils::StringToSoundWave(Data);
    if (ensure(SoundWave))
    {
#if INWORLD_OVR_LIPSYNC
        UOVRLipSyncFrameSequence* LipSyncSequence = nullptr;
        if (LipSyncComponent.Get())
        {
            LipSyncSequence = SoundWaveToLipSyncSequence(SoundWave, (uint8*)Data.data(), Data.size());
        }
#endif

        if (auto* Utterance = FindOrAddUtterance(Event.PacketId.InteractionId, Event.PacketId.UtteranceId))
        {
            Utterance->SoundWave = SoundWave;
#if INWORLD_OVR_LIPSYNC
            Utterance->LipSyncSequence = LipSyncSequence;
#endif
        }
    }
}

void UInworldCharacterComponent::Visit(const Inworld::FControlEvent& Event)
{
    if (Event.GetControlAction() == ai::inworld::packets::ControlEvent_Action_INTERACTION_END)
    {
        PlayerInteractionHistory.ClearCanceledInteraction(Event.PacketId.InteractionId);
		
        if (PendingTrigger.InteractionId == Event.PacketId.InteractionId)
		{
			GetWorld()->GetSubsystem<UInworldApiSubsystem>()->NotifyCustomTrigger(PendingTrigger.Name);
			PendingTrigger.InteractionId = NAME_None;
			PendingTrigger.Name.Empty();
		}
    }
}

void UInworldCharacterComponent::Visit(const Inworld::FEmotionEvent& Event)
{
    EInworldCharacterEmotionalState State = EInworldCharacterEmotionalState::Idle;

    const auto& EmotionState = Event.GetEmotionalState();

    constexpr float Tolerance = 0.1f;
    if (FMath::IsNearlyEqual(EmotionState.Joy, 1.f, Tolerance))
    {
        State = EInworldCharacterEmotionalState::Joy;
    }
    else if (FMath::IsNearlyEqual(EmotionState.Joy, -1.f, Tolerance))
    {
        State = EInworldCharacterEmotionalState::Sadness;
    }
    else if (FMath::IsNearlyEqual(EmotionState.Fear, 1.f, Tolerance))
    {
        State = EInworldCharacterEmotionalState::Fear;
    }
    else if (FMath::IsNearlyEqual(EmotionState.Fear, -1.f, Tolerance))
    {
        State = EInworldCharacterEmotionalState::Anger;
    }
    else if (FMath::IsNearlyEqual(EmotionState.Trust, 1.f, Tolerance))
    {
        State = EInworldCharacterEmotionalState::Trust;
    }
    else if (FMath::IsNearlyEqual(EmotionState.Trust, -1.f, Tolerance))
    {
        State = EInworldCharacterEmotionalState::Disgust;
    }
    else if (FMath::IsNearlyEqual(EmotionState.Surprise, 1.f, Tolerance))
    {
        State = EInworldCharacterEmotionalState::Surprise;
    }
    else if (FMath::IsNearlyEqual(EmotionState.Surprise, -1.f, Tolerance))
    {
        State = EInworldCharacterEmotionalState::Anticipation;
    }

    if (EmotionalState != State)
    {
        EmotionalState = State;
        OnEmotionalStateChanged.Broadcast(EmotionalState);
        CurrentEmotionMontages.Empty();
    }

    const auto Behavior = static_cast<EInworldCharacterEmotionalBehavior>(Event.GetEmotionalBehavior());
    if (Behavior != EmotionalBehavior)
    {
        EmotionalBehavior = Behavior;
        OnEmotionalBehaviorChanged.Broadcast(EmotionalBehavior);
    }
}

void UInworldCharacterComponent::Visit(const Inworld::FCustomEvent& Event)
{
    if (IsCustomGesture(Event.GetName()))
    {
        if (auto* Utterance = FindOrAddUtterance(Event.PacketId.InteractionId, Event.PacketId.UtteranceId))
        {
            Utterance->CustomGesture = Event.GetName();
        }
    }
    else
    {
        PendingTrigger.Name = Event.GetName();
        PendingTrigger.InteractionId = Event.PacketId.InteractionId;
    }
}

void FInworldCharacterInteractionHistory::Add(const Inworld::FTextEvent& Event)
{
    int32 Idx;
    if (CanceledInteractions.Find(Event.PacketId.InteractionId, Idx))
    {
        return;
    }

    auto* Interaction = Interactions.FindByPredicate([&Event](const auto& I) { return I.Event.PacketId.UtteranceId == Event.PacketId.UtteranceId; });
    if (Interaction)
    {
        auto Ts = Interaction->Event.Timestamp;
        *Interaction = FInworldCharacterInteraction(Event);
        Interaction->Event.Timestamp = Ts;
        return;
    }

    Interactions.Add(Event);
}

void FInworldCharacterInteractionHistory::Clear()
{
    Interactions.Empty();
}

void FInworldCharacterInteractionHistory::CancelUtterance(const FName& InteractionId, const FName& UtteranceId)
{
    CanceledInteractions.Add(InteractionId);
    Interactions.RemoveAll([&UtteranceId, &InteractionId](const auto& I) { return I.Event.PacketId.UtteranceId == UtteranceId && I.Event.PacketId.InteractionId == InteractionId; });
}

bool FInworldCharacterInteractionHistory::IsInteractionCanceled(const FName& InteractionId) const
{
    int32 Idx;
    return CanceledInteractions.Find(InteractionId, Idx);
}

void FInworldCharacterInteractionHistory::ClearCanceledInteraction(const FName& InteractionId)
{
    CanceledInteractions.Remove(InteractionId);
}

FInworldCharacterInteraction::FInworldCharacterInteraction(const Inworld::FTextEvent& InEvent)
    : Text(InEvent.GetText())
    , bPlayerInteraction(InEvent.Routing.Source.Type == ai::inworld::packets::Actor_Type_PLAYER)
    , Event(InEvent)
{}

bool FInworldCharacterUtterance::HasTextAndAudio() const
{
#if INWORLD_OVR_LIPSYNC
    if (!LipSyncSequence)
    {
        return false;
    }
#endif

	return SoundWave && !TextEvent.GetText().IsEmpty();
}
