#include "InworldPlayerComponent.h"
#include "proto/ProtoDisableWarning.h"
#include "InworldApi.h"
#include "InworldCharacterComponent.h"
#include "AudioCaptureComponent.h"
#include "AudioMixerBlueprintLibrary.h"
#include "AudioMixerDevice.h"
#include "AudioMixerSubmix.h"
#include "InworldUtils.h"

UInworldPlayerComponent::UInworldPlayerComponent()
{
    PrimaryComponentTick.bCanEverTick = true;
    PrimaryComponentTick.bTickEvenWhenPaused = true;
}

void UInworldPlayerComponent::BeginPlay()
{
    Super::BeginPlay();

    InworldSubsystem = GetWorld()->GetSubsystem<UInworldApiSubsystem>();
    InworldSubsystem->RegisterPlayerComponent(this);

    AudioCaptureComponent = Cast<UAudioCaptureComponent>(GetOwner()->GetComponentByClass(UAudioCaptureComponent::StaticClass()));
    if (!ensure(AudioCaptureComponent.Get() && AudioCaptureComponent->SoundSubmix))
    {
        Inworld::Utils::ErrorLog("UInworldPlayerComponent::BeginPlay: AudioCaptureComponent is invalid.", true);
    }
}

void UInworldPlayerComponent::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
    if (!TargetCharacter.IsValid())
    {
        StopVoiceCapture();
    }

    if (InworldSubsystem.IsValid())
    {
        InworldSubsystem->UnregisterPlayerComponent();
    }

    Super::EndPlay(EndPlayReason);
}

void UInworldPlayerComponent::TickComponent(float DeltaTime, enum ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
    Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

    if (TickType == ELevelTick::LEVELTICK_PauseTick)
    {
        StopVoiceCapture();
        return;
    }

    if (InworldSubsystem->GetConnectionState() != EInworldConnectionState::Connected)
	{
        return;
    }

    // restore voice capture after pause
	if (TargetCharacter.IsValid() && !bCapturingVoice)
	{
        StartVoiceCapture();
	}

	UpdateTargetCharacter();
	UpdateVoiceCapture();
}

void UInworldPlayerComponent::HandleConnectionStateChanged(EInworldConnectionState State)
{
    if (State != EInworldConnectionState::Connected)
    {
        StopVoiceCapture();
    }
}

void UInworldPlayerComponent::Visit(const Inworld::FTextEvent& Event)
{
    if (TargetCharacter.IsValid())
    {
        TargetCharacter->HandlePlayerTalking(Event);
    }
}

void UInworldPlayerComponent::StartVoiceCapture()
{
    if (bCapturingVoice)
    {
        return;
    }

    if (!ensure(TargetCharacter.IsValid()))
    {
        Inworld::Utils::ErrorLog("UInworldPlayerComponent::StartVoiceCapture: TargetAgentId is empty.", true);
        return;
    }

	AudioCaptureComponent->Start();
    InworldSubsystem->StartAudioSession(TargetCharacter->GetAgentId());
    StartVoiceChunkCapture();
    bCapturingVoice = true;
}

void UInworldPlayerComponent::StopVoiceCapture()
{
    if (!bCapturingVoice)
    {
        return;
    }

    bCapturingVoice = false;
    StopVoiceChunkCapture();
	AudioCaptureComponent->Stop();
    if (TargetCharacter.IsValid())
    {
        InworldSubsystem->StopAudioSession(TargetCharacter->GetAgentId());
    }
}

void UInworldPlayerComponent::HandlePacket(TSharedPtr<Inworld::FInworldPacket> Packet)
{
    Packet->Accept(*this);
}

void UInworldPlayerComponent::SetCameraForward(const FVector& InCameraForward)
{
    CameraForward = InCameraForward;
}

void UInworldPlayerComponent::StartVoiceChunkCapture()
{
    LastSoundMessageTime = GetWorld()->GetTimeSeconds();
    UAudioMixerBlueprintLibrary::StartRecordingOutput(this, SoundMessageLength, Cast<USoundSubmix>(AudioCaptureComponent->SoundSubmix));
}

void UInworldPlayerComponent::StopVoiceChunkCapture()
{
    LastSoundMessageTime = GetWorld()->GetTimeSeconds();

    auto* SoundWave = UAudioMixerBlueprintLibrary::StopRecordingOutput(this, EAudioRecordingExportType::SoundWave, "InworldVoiceCapture", "InworldVoiceCapture", Cast<USoundSubmix>(AudioCaptureComponent->SoundSubmix));
    if (TargetCharacter.IsValid() && SoundWave)
    {
        InworldSubsystem->SendAudioMessage(TargetCharacter->GetAgentId(), SoundWave);
    }
}

void UInworldPlayerComponent::UpdateTargetCharacter()
{
    const auto& CharacterComponents = InworldSubsystem->GetCharacterComponents();
    TWeakObjectPtr<UInworldCharacterComponent> ClosestCharacter;
    const float MinDistSq = InteractionDistance * InteractionDistance;
    float CurMaxDot = -1.f;
    const FVector Location = GetOwner()->GetActorLocation();
    for (auto& Character : CharacterComponents)
    {
        if (!Character)
        {
            continue;
        }

        const FVector CharacterLocation = Character->GetComponentOwner()->GetActorLocation();
        const float DistSq = FVector::DistSquared(Location, CharacterLocation);
        if (DistSq > MinDistSq)
        {
            continue;
        }

        const FVector2D Direction2D = FVector2D(CharacterLocation - Location).GetSafeNormal();
        const FVector2D Forward2D = FVector2D(CameraForward);
        const float Dot = FVector2D::DotProduct(Forward2D, Direction2D);
        if (Dot < InteractionDotThreshold)
        {
            continue;
        }

        if (Dot < CurMaxDot)
        {
            continue;
        }

        ClosestCharacter = static_cast<UInworldCharacterComponent*>(Character);
        CurMaxDot = Dot;
    }

    if (TargetCharacter.IsValid() && (!ClosestCharacter.IsValid() || TargetCharacter.Get() != ClosestCharacter.Get()))
    {
        ClearTargetCharacter();
    }

    if (ClosestCharacter.IsValid() && !TargetCharacter.IsValid())
    {
        SetTargetCharacter(ClosestCharacter);
    }
}

void UInworldPlayerComponent::SetTargetCharacter(TWeakObjectPtr<UInworldCharacterComponent> ClosestCharacter)
{
    TargetCharacter = ClosestCharacter;
    TargetCharacter->HandlePlayerInteraction(true);
    StartVoiceCapture();
    Inworld::Utils::Log(FString::Printf(TEXT("Player started interacting with: %s"), *TargetCharacter->GetUiName()), true, Inworld::Utils::FTextColors::Player());
}

void UInworldPlayerComponent::ClearTargetCharacter()
{
    if (TargetCharacter.IsValid())
    {
        StopVoiceCapture();
        TargetCharacter->HandlePlayerInteraction(false);
        Inworld::Utils::Log(FString::Printf(TEXT("Player stopped interacting with: %s"), *TargetCharacter->GetUiName()), true, Inworld::Utils::FTextColors::Player());
        TargetCharacter = nullptr;
    }
}

void UInworldPlayerComponent::UpdateVoiceCapture()
{
    if (bCapturingVoice && GetWorld()->GetTimeSeconds() > LastSoundMessageTime + SoundMessageLength)
    {
        StopVoiceChunkCapture();
        StartVoiceChunkCapture();
    }
}

