#pragma once


#include "CoreMinimal.h"
#include "InworldClient.h"
#include "Subsystems/WorldSubsystem.h"
#include "InworldState.h"
#include "InworldComponentInterface.h"
#include "InworldApi.generated.h"

namespace Inworld
{
	class ICharacterComponent;
	class IPlayerComponent;
}
class USoundWave;

DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FOnConnectionStateChanged, EInworldConnectionState, State);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FCustomTrigger, FString, Name);

UCLASS(BlueprintType, Config = Engine)
class INWORLDAIINTEGRATION_API UInworldApiSubsystem : public UTickableWorldSubsystem
{
	GENERATED_BODY()

public:
    /**
     * Start InworldAI session
     * call after all Player/Character components have been registered
     * @param SceneName : full inworld studio scene name
     */
    UFUNCTION(BlueprintCallable, Category = "Inworld")
    void StartSession(const FString& SceneName, const FString& PlayerName);

    /**
     * Register Character component
     * call before StartSession
     */
    void RegisterCharacterComponent(Inworld::ICharacterComponent* Component);
    void UnregisterCharacterComponent(Inworld::ICharacterComponent* Component);

    /**
     * Register Player component
     * call before StartSession
     */
	void RegisterPlayerComponent(Inworld::IPlayerComponent* Component);
	void UnregisterPlayerComponent();

    /** Send text to agent */
	UFUNCTION(BlueprintCallable, Category = "Messages")
    void SendTextMessage(const FName& AgentId, const FString& Text);

    /** Send custom event to agent */
	UFUNCTION(BlueprintCallable, Category = "Messages")
	void SendCustomEvent(FName AgentId, const FString& Name);

    /** Send custom event to player's current target agent */
	UFUNCTION(BlueprintCallable, Category = "Messages")
	void SendCustomEventForTargetAgent(const FString& Name);

    /**
     * Send audio to agent
     * start audio session before sending audio
     * stop audio session after all audio chunks have been sent
     * chunks should be ~100ms
     */
    UFUNCTION(BlueprintCallable, Category = "Messages")
    void SendAudioMessage(const FName& AgentId, USoundWave* SoundWave);
    
    /**
     * Start audio session with agent
     * call before sending audio messages
     */
    UFUNCTION(BlueprintCallable, Category = "Audio")
    void StartAudioSession(const FName& AgentId);

    /**
     * Stop audio session with agent
     * call after all audio messages have been sent
     */
    UFUNCTION(BlueprintCallable, Category = "Audio")
    void StopAudioSession(const FName& AgentId);

    /** Get current connection state */
    UFUNCTION(BlueprintCallable, Category = "Connection")
	EInworldConnectionState GetConnectionState() const { return Client.GetConnectionState(); }
    
    /** Get all registered character components */
	const TArray<Inworld::ICharacterComponent*>& GetCharacterComponents() const { return CharacterComponentRegistry; }

    /** Cancel agents response in case agent has been interrupted by player */
    UFUNCTION(BlueprintCallable, Category = "Messages")
    void CancelResponse(const FName& AgentId, const FName& InteractionId, const TArray<FName>& UtteranceIds);

    /** 
    * Call on Inworld::FCustomEvent coming to agent
    * custom events meant to be triggered on interaction end (see InworldCharacterComponent)
    */
    void NotifyCustomTrigger(const FString& Name) { OnCustomTrigger.Broadcast(Name); }

    /** Subsystem interface */
    virtual bool DoesSupportWorldType(EWorldType::Type WorldType) const override;
    virtual void Tick(float DeltaTime) override;
    virtual TStatId GetStatId() const override { RETURN_QUICK_DECLARE_CYCLE_STAT(UInworldApiSubsystem, STATGROUP_Tickables); }
    virtual void Deinitialize();

    UPROPERTY(BlueprintAssignable, BlueprintCallable, Category = "EventDispatchers")
    FOnConnectionStateChanged OnConnectionStateChanged;

    UPROPERTY(BlueprintAssignable, BlueprintCallable, Category = "EventDispatchers")
    FCustomTrigger OnCustomTrigger;

private:
    UPROPERTY(EditAnywhere, config, Category = "Connection")
	FString AuthUrl;

	UPROPERTY(EditAnywhere, config, Category = "Connection")
	FString TargetUrl;

	UPROPERTY(EditAnywhere, config, Category = "Connection")
	FString ApiKey;

	UPROPERTY(EditAnywhere, config, Category = "Connection")
	FString ApiSecret;

    TMap<FName, Inworld::ICharacterComponent*> CharacterComponentByAgentId;
    TArray<Inworld::ICharacterComponent*> CharacterComponentRegistry;

    Inworld::IPlayerComponent* PlayerComponent;

    Inworld::FClient Client;
};