#pragma once

#include "CoreMinimal.h"
#include "InworldState.h"
#include "Packets.h"
#include "Engine/DataTable.h"
#include "Components/ActorComponent.h"
#include "InworldComponentInterface.h"

#include "InworldCharacterComponent.generated.h"

#if INWORLD_OVR_LIPSYNC
class UOVRLipSyncFrameSequence;
class UOVRLipSyncPlaybackActorComponent;
#endif

struct FActorComponentTickFunction;
class UAudioComponent;
class USoundWave;

USTRUCT()
struct FInworldCharacterUtterance
{
    GENERATED_BODY();

    bool IsValid() const { return HasTextAndAudio() || (!CustomGesture.IsEmpty() && !SoundWave && TextEvent.GetText().IsEmpty()); }
    bool HasTextAndAudio() const;

    UPROPERTY()
    USoundWave* SoundWave = nullptr;

	UPROPERTY()
    UObject* LipSyncSequence = nullptr;

    FString CustomGesture;

    FName UtteranceId;
    FName InteractionId;
    Inworld::FTextEvent TextEvent;
};

UENUM(BlueprintType)
enum class EInworldCharacterState : uint8
{
    Idle,
    Interested,
    Listening,
    Thinking,
    Talking
};

UENUM(BlueprintType)
enum class EInworldCharacterSimpleGesture : uint8
{
    Unknown = 0,
    Wave = 1,
    ThumbUp = 2,
    ThumbDown = 3,
    Happy = 4,
    Sad = 5,
    Surprised = 6,
};

UENUM(BlueprintType)
enum class EInworldCharacterEmotionalState : uint8
{
    Idle,
    Joy,
    Sadness,
    Fear,
    Anger,
    Disgust,
    Trust,
    Anticipation,
    Surprise,
};

UENUM(BlueprintType)
enum class EInworldCharacterEmotionalBehavior : uint8
{
	NEUTRAL = 0,
	DISGUST = 1,
	CONTEMPT = 2,
	BELLIGERENCE = 3,
	DOMINEERING = 4,
	CRITICISM = 5,
	ANGER = 6,
	TENSION = 7,
	TENSE_HUMOR = 8,
	DEFENSIVENESS = 9,
	WHINING = 10,
	SADNESS = 11,
	STONEWALLING = 12,
	INTEREST = 13,
	VALIDATION = 14,
	AFFECTION = 15,
	HUMOR = 16,
	SURPRISE = 17,
	JOY = 18,
};

USTRUCT(BlueprintType)
struct INWORLDAIINTEGRATION_API FInworldAnimationTableRow : public FTableRowBase
{
    GENERATED_BODY()
            
public:
    UPROPERTY(BlueprintReadOnly, EditDefaultsOnly, Category = "Animation", DisplayName = "Emotion")
    EInworldCharacterEmotionalBehavior Key;

    UPROPERTY(BlueprintReadOnly, EditDefaultsOnly, Category = "Animation")
    TArray<UAnimMontage*> Montages;
};

USTRUCT(BlueprintType)
struct INWORLDAIINTEGRATION_API FInworldSemanticGestureTableRow : public FTableRowBase
{
	GENERATED_BODY()

public:
	UPROPERTY(BlueprintReadOnly, EditDefaultsOnly, Category = "Animation", DisplayName = "Semantics")
    FString Key;

	UPROPERTY(BlueprintReadOnly, EditDefaultsOnly, Category = "Animation")
	TArray<UAnimMontage*> Montages;
};

USTRUCT(BlueprintType)
struct FInworldCharacterInteraction
{
    GENERATED_BODY();

    FInworldCharacterInteraction() = default;
    FInworldCharacterInteraction(const Inworld::FTextEvent& InEvent);

    UPROPERTY(BlueprintReadOnly, Category = "Text")
    FString Text;
    
    UPROPERTY(BlueprintReadOnly, Category = "Interaction")
    bool bPlayerInteraction = false;

    Inworld::FTextEvent Event;
};

USTRUCT(BlueprintType)
struct FInworldCharacterInteractionHistory
{
    GENERATED_BODY();

    void Add(const Inworld::FTextEvent& Event);
    void Clear();

    TArray<FInworldCharacterInteraction>& GetInteractions() { return Interactions; }

    void CancelUtterance(const FName& InteractionId, const FName& UtteranceId);
    bool IsInteractionCanceled(const FName& InteractionId) const;
    void ClearCanceledInteraction(const FName& InteractionId);

private:
    TArray<FInworldCharacterInteraction> Interactions;
    TArray<FName> CanceledInteractions;
};

DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FInworldCharacterPlayerInteractionStateChanged, bool, bInteracting);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FOnInworldCharacterStateChanged, EInworldCharacterState, State);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FOnInworldCharacterSimpleGesture, EInworldCharacterSimpleGesture, Gesture);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FOnInworldCharacterCustomGesture, FString, Gesture);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FOnInworldCharacterCustomEvent, FString, Name);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FOnInworldCharacterEmotionalStateChanged, EInworldCharacterEmotionalState, EmotionalState);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FOnInworldCharacterEmotionalBehaviorChanged, EInworldCharacterEmotionalBehavior, EmotionalBehavior);
DECLARE_DYNAMIC_MULTICAST_DELEGATE_OneParam(FOnInworldCharacterInteractionsChanged, const TArray<FInworldCharacterInteraction>&, Interactions);

DECLARE_DYNAMIC_MULTICAST_DELEGATE_TwoParams(FOnInworldCharacterUtteranceStarted, float, Duration, FString, CustomGesture);
DECLARE_DYNAMIC_MULTICAST_DELEGATE(FOnInworldCharacterUtteranceStopped);
DECLARE_DYNAMIC_MULTICAST_DELEGATE(FOnInworldCharacterUtteranceInterrupted);

UCLASS(ClassGroup = (Custom), meta = (BlueprintSpawnableComponent))
class INWORLDAIINTEGRATION_API UInworldCharacterComponent : public UActorComponent, public Inworld::FPacketVisitor, public Inworld::ICharacterComponent
{
	GENERATED_BODY()

public:
	UInworldCharacterComponent();

    UPROPERTY(BlueprintAssignable, Category = "EventDispatchers")
    FInworldCharacterPlayerInteractionStateChanged OnPlayerInteractionStateChanged;

    UPROPERTY(BlueprintAssignable, Category = "EventDispatchers")
    FOnInworldCharacterStateChanged OnStateChanged;

    UPROPERTY(BlueprintAssignable, Category = "EventDispatchers")
	FOnInworldCharacterEmotionalStateChanged OnEmotionalStateChanged;

	UPROPERTY(BlueprintAssignable, Category = "EventDispatchers")
	FOnInworldCharacterEmotionalBehaviorChanged OnEmotionalBehaviorChanged;

    UPROPERTY(BlueprintAssignable, Category = "EventDispatchers")
    FOnInworldCharacterInteractionsChanged OnInteractionsChanged;

    UPROPERTY(BlueprintAssignable, Category = "EventDispatchers")
    FOnInworldCharacterUtteranceStarted OnUtteranceStarted;

    UPROPERTY(BlueprintAssignable, Category = "EventDispatchers")
    FOnInworldCharacterUtteranceStopped OnUtteranceStopped;

    UPROPERTY(BlueprintAssignable, Category = "EventDispatchers")
    FOnInworldCharacterUtteranceInterrupted OnUtteranceInterrupted;

    virtual void BeginPlay() override;
    virtual void EndPlay(EEndPlayReason::Type Reason);
	virtual void TickComponent(float DeltaTime, enum ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;
    
    UFUNCTION(BlueprintCallable, Category = "Inworld")
	virtual const FName& GetAgentId() const override { return AgentId; }
    virtual void SetAgentId(const FName& InAgentId) override { AgentId = InAgentId; }

    virtual AActor* GetComponentOwner() const override { return GetOwner(); }

    const FString& GetUiName() const { return UiName; }
    void SetUiName(const FString& Name) { UiName = Name; }

    void SendTextMessage(const FString& Text);

    virtual const FString& GetBrainName() const override { return BrainName; }

    virtual void HandlePacket(TSharedPtr<Inworld::FInworldPacket> Packet) override;

    virtual void HandleConnectionStateChanged(EInworldConnectionState State) override {}

    void HandlePlayerInteraction(bool bInteracting);
    void HandlePlayerTalking(const Inworld::FTextEvent& Event);

    UFUNCTION(BlueprintCallable, Category = "Emotions")
	EInworldCharacterEmotionalState GetEmotionalState() const { return EmotionalState; }

	UFUNCTION(BlueprintCallable, Category = "Emotions")
	EInworldCharacterEmotionalBehavior GetEmotionalBehavior() const { return EmotionalBehavior; }

    UFUNCTION(BlueprintCallable, Category = "Interactions")
    const TArray<FInworldCharacterInteraction>& GetPlayerInteractions() { return PlayerInteractionHistory.GetInteractions(); }

    UFUNCTION(BlueprintCallable, Category = "Interactions")
    bool IsInteractingWithPlayer() const { return bInteractingWithPlayer; }

    UFUNCTION(BlueprintCallable, Category = "Animation")
    UAnimMontage* GetMontageForCurrentEmotion(float UtteranceDuration) const { return GetMontageForEmotion(EmotionalBehavior, UtteranceDuration); }

    UFUNCTION(BlueprintCallable, Category = "Animation")
	UAnimMontage* GetMontageForEmotion(EInworldCharacterEmotionalBehavior State, float UtteranceDuration) const;
		
    UFUNCTION(BlueprintCallable, Category = "Animation")
	UAnimMontage* GetMontageForCustomGesture(const FString& Semantic, float UtteranceDuration) const;

    UFUNCTION(BlueprintCallable, Category = "Animation")
	float GetRemainingTimeForCurrentUtterance() const;

	UFUNCTION(BlueprintCallable, Category = "Events")
	void SendCustomEvent(const FString& Name) const;

protected:
    UPROPERTY(BlueprintReadOnly, EditAnywhere, Category = "Animation")
	UDataTable* AnimationDT;

	UPROPERTY(BlueprintReadOnly, EditAnywhere, Category = "Animation")
	UDataTable* SemanticGestureDT;

    UPROPERTY(BlueprintReadOnly, Category = "State")
    EInworldCharacterState CurrentState = EInworldCharacterState::Idle;

    UPROPERTY(EditAnywhere, BlueprintReadOnly, Category = "UI")
    FString UiName = "Character";

private:
    struct FPendingTrigger
    {
		FName InteractionId;
        FString Name;
    };

    virtual void Visit(const Inworld::FTextEvent& Event) override;
    virtual void Visit(const Inworld::FDataEvent& Event) override;
    virtual void Visit(const Inworld::FControlEvent& Event) override;
    virtual void Visit(const Inworld::FEmotionEvent& Event) override;
    virtual void Visit(const Inworld::FCustomEvent& Event) override;

    void SetState(EInworldCharacterState State);

    FInworldCharacterUtterance* FindOrAddUtterance(const FName& InteractionId, const FName& UtteranceId);

    void OnAudioPlaybackPercent(const UAudioComponent* InAudioComponent, const USoundWave* InSoundWave, float Percent);

	bool IsCustomGesture(const FString& CustomEventName) const;

#if INWORLD_OVR_LIPSYNC
	UOVRLipSyncFrameSequence* SoundWaveToLipSyncSequence(USoundWave* SoundWave, uint8* Data, int32 Num);
#endif

    UPROPERTY(EditAnywhere, Category = "Inworld")
    FString BrainName;

	UPROPERTY()
	TArray<FInworldCharacterUtterance> PendingUtterances;

    UPROPERTY()
    mutable TArray<UAnimMontage*> CurrentEmotionMontages;

    FInworldCharacterInteractionHistory PlayerInteractionHistory;
    
    FDelegateHandle AudioPlaybackPercentHandle;

    // interaction played currently on client
    FName CurrentInteractionId = NAME_None;
    FName CurrentUtteranceId = NAME_None;

    FPendingTrigger PendingTrigger;

    EInworldCharacterEmotionalState EmotionalState = EInworldCharacterEmotionalState::Idle;
    EInworldCharacterEmotionalBehavior EmotionalBehavior = EInworldCharacterEmotionalBehavior::NEUTRAL;

    TWeakObjectPtr<UAudioComponent> AudioComponent;

#if INWORLD_OVR_LIPSYNC
    TWeakObjectPtr<UOVRLipSyncPlaybackActorComponent> LipSyncComponent;
#endif

    FName AgentId = NAME_None;

    float CurrentAudioPlaybackPercent = 0.f;

    bool bInteractingWithPlayer = false;
    bool bListeningToPlayer = false;
    bool bPendingResponseToPLayer = false;
};
